/****************************************************************************
 * Functions to load fonts.
 * Copyright (C) 2004 Joe Wingbermuehle
 ****************************************************************************/

#include "jwm.h"

static const char *DEFAULT_FONT = "5x7";

static char *fontNames[FONT_COUNT] = { NULL };
static int antialias[FONT_COUNT] = { 0 };
XFontStruct *fonts[FONT_COUNT] = { NULL };

/* Data for antialiasing */
static unsigned long *inputData = NULL;
static unsigned long *outputData = NULL;
static XImage *inputImage = NULL;
static XImage *outputImage = NULL;
static Pixmap buffer = None;
static GC bufferGC;

static unsigned long white;
static unsigned long black;

static void Antialias(int width, int height, RampType ramp);

/****************************************************************************
 ****************************************************************************/
void InitializeFonts() {

	int x;

	white = WhitePixel(display, rootScreen);
	black = BlackPixel(display, rootScreen);

	for(x = 0; x < FONT_COUNT; x++) {
		fonts[x] = JXLoadQueryFont(display, fontNames[x]);
		if(!fonts[x]) {
			fonts[x] = JXLoadQueryFont(display, DEFAULT_FONT);
			if(!fonts[x]) {
				FatalError("Could not load the default font: %s", DEFAULT_FONT);
			}
		}
	}

}

/****************************************************************************
 ****************************************************************************/
void SetFont(FontType type, const char *value, int aa) {

	Assert(value);

	if(fontNames[type]) {
		Release(fontNames[type]);
	}

	antialias[type] = aa;
	fontNames[type] = Allocate(strlen(value) + 1);
	strcpy(fontNames[type], value);

}

/****************************************************************************
 ****************************************************************************/
void Antialias(int width, int height, RampType ramp) {

	int x, y;
	int offset;
	int value;
	int line = inputImage->width;

	for(y = 1; y < height - 1; y++) {
		for(x = 1; x < width - 1; x++) {

			offset = y * line + x;

			if(   inputData[offset + 1] != inputData[offset - 1]
				&& inputData[offset + line + 1] != inputData[offset - line - 1]
				&& inputData[offset + line - 1] != inputData[offset - line + 1]
				&& inputData[offset + line] != inputData[offset - line]) {

				if(inputData[offset] == black) {
					value = 2;
				} else {
					value = 0;
				}
				if(inputData[offset + 1] == black) {
					++value;
				}
				if(inputData[offset - 1] == black) {
					++value;
				}
				if(inputData[offset + line] == black) {
					++value;
				}
				if(inputData[offset - line] == black) {
					++value;
				}

				outputData[offset] = ramps[ramp][value];

			} else {
				if(inputData[offset] == black) {
					outputData[offset] = ramps[ramp][7];
				} else {
					outputData[offset] = ramps[ramp][0];
				}
			}

		}
	}

}

/****************************************************************************
 ****************************************************************************/
void RenderString(Drawable d, GC g, FontType font, RampType ramp,
	int x, int y, int width, const char *str) {

	int len, h, w;

	Assert(str);

	len = strlen(str);

	if(!antialias[font]) {
		switch(ramp) {
		case RAMP_BORDER:
			JXSetForeground(display, g, colors[COLOR_BORDER_FG]);
			break;
		case RAMP_BORDER_ACTIVE:
			JXSetForeground(display, g, colors[COLOR_BORDER_ACTIVE_FG]);
			break;
		case RAMP_TRAY:
			JXSetForeground(display, g, colors[COLOR_TRAY_FG]);
			break;
		case RAMP_TRAY_ACTIVE:
			JXSetForeground(display, g, colors[COLOR_TRAY_ACTIVE_FG]);
			break;
		case RAMP_MENU:
			JXSetForeground(display, g, colors[COLOR_MENU_FG]);
			break;
		case RAMP_MENU_ACTIVE:
			JXSetForeground(display, g, colors[COLOR_MENU_ACTIVE_FG]);
			break;
		default:
			break;
		}
		JXSetFont(display, g, fonts[font]->fid);
		JXDrawString(display, d, g, x, y + fonts[font]->ascent, str, len);
		return;
	}

	h = Min(fonts[font]->ascent + fonts[font]->descent, rootHeight) + 2;
	w = Min(JXTextWidth(fonts[font], str, len), width) + 2;

	if(!inputData || h > inputImage->height || w > inputImage->width) {

		if(inputData) {

			Release(inputData);
			inputImage->data = NULL;
			JXDestroyImage(inputImage);

			Release(outputData);
			outputImage->data = NULL;
			JXDestroyImage(outputImage);

			JXFreeGC(display, bufferGC);
			JXFreePixmap(display, buffer);

		}

		buffer = JXCreatePixmap(display, rootWindow, w, h, rootDepth);
		bufferGC = JXCreateGC(display, buffer, 0, NULL);

		inputData = Allocate(w * h * sizeof(unsigned long));
		inputImage = JXCreateImage(display,
			DefaultVisual(display, rootScreen), rootDepth,
			ZPixmap, 0, (char*)inputData, w, h, 32, 0);

		outputData = Allocate(w * h * sizeof(unsigned long));
		outputImage = JXCreateImage(display,
			DefaultVisual(display, rootScreen), rootDepth,
			ZPixmap, 0, (char*)outputData, w, h, 32, 0);

	}

	JXSetForeground(display, bufferGC, white);
	JXFillRectangle(display, buffer, bufferGC, 0, 0, w, h);

	JXSetForeground(display, bufferGC, black);
	JXSetFont(display, bufferGC, fonts[font]->fid);
	JXDrawString(display, buffer, bufferGC, 1,
		1 + fonts[font]->ascent, str, len);

	JXGetSubImage(display, buffer, 0, 0, w, h, AllPlanes, ZPixmap,
		inputImage, 0, 0);

	Antialias(w, h, ramp);

	JXPutImage(display, d, g, outputImage, 1, 1, x, y, w - 2, h - 2);

}

/****************************************************************************
 ****************************************************************************/
void ReleaseFonts() {
	int x;

	for(x = 0; x < FONT_COUNT; x++) {
		if(fonts[x]) {
			JXFreeFont(display, fonts[x]);
			fonts[x] = NULL;
		}
		if(fontNames[x]) {
			Release(fontNames[x]);
			fontNames[x] = NULL;
		}
	}

	if(inputData) {
		Release(inputData);
		inputImage->data = NULL;
		JXDestroyImage(inputImage);

		Release(outputData);
		outputImage->data = NULL;
		JXDestroyImage(outputImage);

		JXFreeGC(display, bufferGC);
		JXFreePixmap(display, buffer);
		buffer = None;
		inputData = NULL;
	}

}


